#include <windows.h>
#include <malloc.h>
#pragma hdrstop

#include "FakeCall.h"
#include "pilotcpu\dragonball.h"

extern void StartDebugger(); // this might come from any debugger

#if DBG_TRACE
	extern Debugger * gpdbg;
#endif // DBG_TRACE

FakeCall::FakeCall()
{
	m_CpuWasRunning = false;

	if ( cpuRunning == CPU_getstate())
	{
		m_CpuWasRunning = true;
#if DBG_TRACE
		gpdbg->printf( "Fak: Stopping CPU\n" );
#endif // DBG_TRACE
		CPU_stop();
	}

	CPU_getregs(origregs);
	regs = origregs;
	SaveMemory = malloc(4096);
	CopyMemory(SaveMemory, CPU_getmemptr(addrstart), 4096);
	nextaddr = addrstart;
}

FakeCall::~FakeCall()
{
	CPU_setregs(origregs);
	CopyMemory(CPU_getmemptr(addrstart), SaveMemory, 4096);
	free(SaveMemory);
}

void FakeCall::PushWord(WORD x)
{
  regs.a[7] -= 2;
  CPU_putword(regs.a[7], x);
}

void FakeCall::PushLong(DWORD x)
{
  regs.a[7] -= 4;
  CPU_putlong(regs.a[7], x);
}

void *FakeCall::PushPtr(int size)
{
  regs.a[7] -= 4;
  if (size == 0) {
    CPU_putlong(regs.a[7], 0);
    return NULL;
  }
  CPU_putlong(regs.a[7], nextaddr);
  void *r = CPU_getmemptr(nextaddr);
  nextaddr += size + (size & 1);
  return r;
}

void FakeCall::Call(WORD api, BOOL debug)
{
	SwapRange(addrstart, nextaddr);

	WORD oldins = CPU_getword(regs.pc);
	WORD oldapi = CPU_getword(regs.pc+2);
	WORD oldnxt = CPU_getword(regs.pc+4);
	PutSwap(regs.pc,0x4E4F);	// TRAP(0xF) = SysTrap
	PutSwap(regs.pc+2, api);
	PutSwap(regs.pc+4, 0x4AFC);	// The ILLEGAL instruction - i.e., breakpoint
	CPU_setregs(regs);
/*
	if (debug) {
		StartDebugger();
		Sleep(INFINITE);
	}
*/
#if DBG_TRACE
	gpdbg->printf( "Fak: Starting CPU on fake code.\n" );
#endif // DBG_TRACE
	CPU_run();
#if DBG_TRACE
	gpdbg->printf( "Fak: Waiting to hit breakpoint.\n" );
#endif // DBG_TRACE
	CPU_wait();
#if DBG_TRACE
	gpdbg->printf( "Fak: CPU stopped.\n" );
#endif // DBG_TRACE
	CPU_getregs(regs);
	CPU_setregs(origregs);
	PutSwap(origregs.pc, oldins);
	PutSwap(origregs.pc+2, oldapi);
	PutSwap(origregs.pc+4, oldnxt);
	SwapRange(addrstart, nextaddr);

	if ( m_CpuWasRunning )
	{
#if DBG_TRACE
		gpdbg->printf( "Fak: Restarting CPU\n" );
#endif // DBG_TRACE
		CPU_run();
	}
}

void FakeCall::PutSwap(DWORD addr, WORD word)
{
  WORD *p = (WORD *)CPU_getmemptr(addr);
  *p = word;
  if (CPU_getword(addr) != word) {
    *p = swap(word);
  }
}

void FakeCall::SwapRange(DWORD startaddr, DWORD endaddr)
{
  WORD *p = (WORD *)CPU_getmemptr(startaddr);
  WORD *e = (WORD *)CPU_getmemptr(endaddr);
  while (p < e) {
    *p = swap(*p);
    p++;
  }
}

DWORD FakeCall::GetResultD0()
{
  return regs.d[0];
}

DWORD FakeCall::GetResultA0()
{
  return regs.a[0];
}

